package be.seeseemelk.llamascript;

import com.fasterxml.jackson.databind.ObjectMapper;
import io.github.ollama4j.OllamaAPI;
import io.github.ollama4j.exceptions.OllamaBaseException;
import io.github.ollama4j.models.chat.OllamaChatMessage;
import io.github.ollama4j.models.chat.OllamaChatMessageRole;
import io.github.ollama4j.models.chat.OllamaChatRequest;
import io.github.ollama4j.utils.OptionsBuilder;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

@Slf4j
@RequiredArgsConstructor
public class LlamaScript {
    private final OllamaAPI api;
    private final String model;
    private final ObjectMapper mapper = new ObjectMapper();

    public <T> T eval(String prompt, Class<T> returnType, Object... arguments) {
        var argumentMap = new HashMap<Integer, Object>();
        for (Object argument : arguments) {
            argumentMap.put(argumentMap.size(), argument);
        }

        if (returnType.equals(Integer.class)) {
            //noinspection unchecked
            return (T) evalInt(prompt, argumentMap);
        } else if (returnType.equals(String.class)) {
            //noinspection unchecked
            return (T) evalString(prompt, argumentMap);
        } else {
            throw new LLamaScriptException("Unsupported return type");
        }
    }

    private String evalString(String prompt, HashMap<Integer, Object> argumentMap) {
        var result = eval(prompt, primitiveOutputType("a string"), argumentMap);
        if (result instanceof String string)
            return string;
        else if (result != null)
            return result.toString();
        else
            throw new LLamaScriptException("No response");
    }

    private Integer evalInt(String prompt, HashMap<Integer, Object> argumentMap) {
        var result = eval(prompt, primitiveOutputType("an integer number"), argumentMap);
        if (result instanceof Integer integer)
            return integer;
        else if (result instanceof String string) {
            try {
                return Integer.parseInt(string);
            } catch (NumberFormatException e) {
                throw new LLamaScriptException("Invalid integer", e);
            }
        }
        else
            throw new LLamaScriptException("Invalid integer type");
    }

    private Object eval(String prompt, String outputType, HashMap<Integer, Object> argumentMap) {
        try {
            var optionsBuilder = new OptionsBuilder();
            optionsBuilder.setSeed(20); // Chosen with 1d20. Yes, a nat 20! Let's go baby!
            optionsBuilder.setTemperature(0f);
            var options = optionsBuilder.build();
            var fullPrompt = new StringBuilder("""
                You will receive a prompt, and will have to consider the question in the prompt with regards to the inputs you will receive.
                Your output must be a JSON object containing the single value 'result'.
                %s
                
                The prompt: %s
                
                Arguments:
                """.formatted(outputType, prompt));
            for (var entry : argumentMap.entrySet()) {
                fullPrompt.append(" - ").append(entry.getKey()).append(": ").append(entry.getValue()).append("\n");
            }
            System.out.printf("Prompt is: %s\n", fullPrompt);

            var messages = new ArrayList<OllamaChatMessage>();
            messages.add(new OllamaChatMessage(OllamaChatMessageRole.USER, fullPrompt.toString()));
            var chatRequest = new OllamaChatRequest(model, messages);
            chatRequest.setReturnFormatJson(true);
            chatRequest.setOptions(options.getOptionsMap());

            var result = api.chat(chatRequest);
            System.out.printf("Result is: %s\n", result.getResponse());
            return mapper.readValue(result.getResponse(), Result.class).result;
        } catch (OllamaBaseException | InterruptedException | IOException e) {
            throw new LLamaScriptException(e);
        }
    }

    private static class Result {
        public Object result;
    }

    private static String primitiveOutputType(String description) {
        return "This result must contain %s.".formatted(description);
    }
}
